%{
/* ----------------------------- Global Definitions ------------------------ */
/* ------------------------------------------------------------------------- */

#undef   input

#include <stdio.h>
#include <dos.h>
#include <ctype.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "ncform.h"
#include "stdhdr.h"

int   start     = FALSE;
int   Syntax    = FALSE;
int   V3Item    = FALSE;
int   V3Count   = 0;
int   Total     = 0;
int   ItemCount = 0;                    /* Number of items in input file.    */
char  *End;

char  Name[64];

int main (int, char **);
void  SelectOutput (void);
unsigned int Hash  (char *);
void InitOutput (void);
void GrowPSP (int);
void InitDescList(void);
char * GetDescription (unsigned, char *);

typedef struct {
   char     *Name;
   char     *String;
   unsigned HashCode;
   FILE     *File;
} SECTION;

/*
** The following type definition is used to allow the user to define the short
** description that will follow the topic name when the list of available
** topics is displayed.  Be advised that the number of entries subortinate to
** a sungle menu item in the Norton Guides determines the ammout of text that
** can be provided for the short descriptions.  Do to the fact that the
** FUNCTIONS list has over 500 function calls documented, short descriptions
** cannot be readily provided due to memory problems when attempting to compile
** the resulting Norton Guide database.  I have provided a TOPICS list for the
** window messages (see file MESSAGES.DSC).  The first element in this file
** is the number of entries in the file (i.e. # of lines), followed by the
** item name and short description for that item.
*/
typedef struct {
   char     *ItemName;                  /* Name of "!short" item.            */
   char     *Header;                    /* Short description to use.         */
   unsigned HashCode;
} HEADERS;
HEADERS  *HdrList;                      /* Allocated array of headers.       */

SECTION  Current;
SECTION  Section[] = {
   { "function.dat",  "AccessResource"},
   { "messages.dat",  "BM_GETCHECK"},
   { "typedefs.dat",  "BITMAP"},
   { "datatype.dat",  "Data Types"},
   { "wndstyle.dat",  "Control Classes"},
   { "function.dat",  "CreateWindowEx"},
   { "macros.dat",    "HIBYTE"},
   { "function.dat",  "HideCaret"},
   { "macros.dat",    "HIWORD"},
   { "function.dat",  "InflateRect"},
   { "macros.dat",    "LOBYTE"},
   { "function.dat",  "LocalAlloc"},
   { "macros.dat",    "LOWORD"},
   { "function.dat",  "LPtoDP"},
   { "macros.dat",    "MAKEINTATOM"},
   { "function.dat",  "MakeProcInstance"},
   { "macros.dat",    "max"},
   { "function.dat",  "MessageBeep"},
   { "macros.dat",    "min"},
   { "function.dat",  "ModifyMenu"},
   { "macros.dat",    "PALETTEINDEX"},
   { "function.dat",  "PatBlt"},
   { "macros.dat",    "RGB"},
   { "function.dat",  "RoundRect"},
   { "mem32bit.dat",  "GetWinMem32Version"},
   { "mem32bit.dat",  "Using the WINMEM32.DLL Library"},
   { "prnescps.dat",  "Printer Escapes"},
   { "fatalxit.dat",  "Fatal Exit Codes"},
   { "messgovr.dat",  "Messages Overview"},
   { "vkeycode.dat",  "Virtual-Key Codes"},
   { "other.dat",     "Data Structures" },
   { "other.dat",     "Microsoft Windows Programmer\'s Reference"},
   { "other.dat",     "Message Functions"},
   { 0, 0, 0, 0 }
};

/* ----------------------------- Begin Productions ------------------------- */
/* ------------------------------------------------------------------------- */
%}

nl    [\n]

%%

^".context" {
   start = TRUE;
}

"----\n\ \n\\b"[^\n]+[\ \n]+"Version [3.0]" {
   start = FALSE;
   V3Item = TRUE;
   End = strtok(&yytext[9], "\\");
   strcpy (Name, &yytext[9]);
   printf ("%04d (V3): ",++Total);
   SelectOutput();
}

"----\n\ \n\\b"[^\n]+ {
   start = FALSE;
   V3Item = FALSE;
   End = strtok(&yytext[9], "\\");
   strcpy (Name, &yytext[9]);
   printf ("%04d (V2): ",++Total);
   SelectOutput();
}

"\\iSyntax" {
   fprintf (yyout, "^BSyntax^B");
   Syntax = TRUE;
}

"\\iReturn Value" {
   fprintf (yyout, "^BReturn^B");
}

"\\iComments" {
   fprintf (yyout, "^BNotes^B ");
}

")\n" {
   Syntax = FALSE;
   fprintf (yyout, ")\n");
}

"\\i"|"\\b"|"\\p" {
   /*
   ** Ignore.
   */
}

[\ \n]+"\-\#\-\n " {
   /*
   ** Ignore.
   */
}

.|\n {
   /*
   ** Output the body of the documentation block, if open.
   */
   if (start == FALSE) {
      putc (yytext[0], yyout);
   }
}

%%

/* -------------------------------- main() --------------------------------- */
/* ------------------------------------------------------------------------- */

PUBLIC main (int argc, char *argv[])
{
   unsigned int    status;
   char            buffer[80];
   char            drive[_MAX_DRIVE];
   char            dir[_MAX_DIR];
   char            fname[_MAX_FNAME];
   char            ext[_MAX_EXT];
   struct find_t   wildbuf;

   if (argc != 2) {
      fprintf (stderr, "NGCVT: Incorrect number of arguments.\n");
      fprintf (stderr, "Usage is: ngcvt <input file>\n");
      exit(1);
   }

   /*
   ** Find the first occurance of the requested file spec (which
   ** can be a wildcard).  Initialize the wildcard structures.
   */
   status = _dos_findfirst (argv[1], _A_NORMAL, &wildbuf);
   if (status) {
      fprintf (stderr, "NGCVT: Cannot find path.\n");
      exit (1);
   }

   _splitpath (argv[1], drive, dir, fname, ext);
   sprintf (buffer, "%s%s%s", drive, dir, wildbuf.name);
   yyin = fopen (buffer, "r");

   /*
   ** Open and initialize the output files.
   */
   GrowPSP (128);
   InitOutput();
   if (yyout == NULL) {
      fprintf (stderr, "NGCVT: Unable to open output file.\n");
      exit (1);
   }

   /*
   ** Initialize the short descriptions list from the input files.
   */
   InitDescList();

   /*
   ** Process each file.
   */
   GrowPSP (128);
   do {
      printf ("%s:\n", buffer);
      if (yyin == NULL) {
         fprintf (stderr, "NGCVT: Error Opening input file.\n");
         exit(1);
      }

      yylex ();
      printf ("\n");

      fclose (yyin);
      yylineno = 0;

   } while (_dos_findnext(&wildbuf) == 0);

   return (0);

}       /* main() */

/* -------------------------------- yywrap() ------------------------------- */
/* ------------------------------------------------------------------------- */

int yywrap ()
{
   return (1);

}

/* -------------------------------- input() -------------------------------- */
/* ------------------------------------------------------------------------- */
/*
**  Name        : input
**
**  Description : This function replaces the default LEX "input" macro with
**                something a little more sophisticated.  The need to do this
**                came about because of the inclusion of non-printable ASCII
**                characters (0xc4 and 0x04) appearing in the code.  This
**                made the LEX program I was using abort.
**
**  Returns     : Input character, as an integer.
**
**  Parameters  : None.
**
*/

int input (void)
{
   if (yysptr > yysbuf) {
      yytchar = U(*--yysptr);
   } else {
      yytchar = getc (yyin);
      if (yytchar == 0xc4) yytchar = (int)'-';
      if (yytchar == 0x04) yytchar = (int)'#';
   }

   if (yytchar == '\n') yylineno++;
   if (yytchar == EOF)  return (0);

   return (yytchar);

}     /* input() */

/* ------------------------------ SelectOutput() --------------------------- */
/* ------------------------------------------------------------------------- */
/*
**  Name        : SelectOutput
**
**  Description : This procedure selects, for each "subject" examined by the
**                lexical component, the appropriate output file.  The text
**                MACROS, and function, and certain tables, are all intermixed.
**                This procedure looks at the title of the subject and, if
**                the title is listed in the "Section" array, executes a
**                transition to a new output file.
**
**                All files are kept open at all times.  This is why I was
**                forced to write the "GrowPSP()" procedure, so that I could
**                be certain I would have enough file handles.
**
**                I use a simple hash code as a preliminary filter and then
**                perform a string comparison to detect collisions.  The
**                handle to the new file is stored in the data record which
**                contains the match.
**
**                If a description is available for the subject, it will be
**                written to the "!short" record in this procedure.
**
**  Returns     : None.
**
**  Parameters  : None.
**
*/

void SelectOutput (void)
{
   unsigned HashVal;
   int      i;
   char     *Description;

   /*
   ** Print the name of the selection.
   */
   printf ("%s\n",Name);

   /*
   ** Check to see if the current string is the delimiter phrase for a
   ** new section.  If so, then change the output file.
   */
   i = 0;
   HashVal = Hash (Name);
   while (Section[i].String != NULL) {
      if (HashVal == Section[i].HashCode) {
         if (strcmp(Name, Section[i].String) == 0) {
            Current = Section[i];
            break;
         }
      }
      i++;
   }

   yyout = Current.File;

   if ((Description = GetDescription(HashVal, Name)) != NULL) {
      fprintf (yyout, "\n!short:%-23s%s\n", Name, Description);
   } else {
      fprintf (yyout, "\n!short:%s\n", Name);
   }
   fprintf (yyout, "^BName^B    %s", Name);
   if (V3Item == TRUE) {
      fprintf (yyout, "\n\n        ^BVersion [3.0]^B\n");
   }

}     /* SelectOutput() */

/* ------------------------------ InitOutput() ----------------------------- */
/* ------------------------------------------------------------------------- */
/*
**  Name        : InitOutput
**
**  Description : Open all of the files defined in the "Section" array.  Be
**                careful not to attempt to open a file twice.  Save the file
**                handle in the "File" memot of the SECTION structure.
**
**  Returns     : None.
**
**  Parameters  : None.
**
**  Side Effects: The "File" member of the "Section" array is initialized.
**                The initial output file is assigned to the lex variable
**                "yyout".
**
*/

void InitOutput (void)
{
   int   i,j;

   i = 0;
   while (Section[i].String != NULL) {
      /*
      ** check to see if the file has been opened before.
      */
      for (j=0;j<i;j++) {
         if (strcmp(Section[j].Name,Section[i].Name) == 0) {
            Section[i].File = Section[j].File;
            break;
         }
      }

      /*
      ** If the current file pointer is NULL, the file has not yet been
      ** opened.  Attempt to open the file and report an error on failure.
      */
      if (Section[i].File == NULL) {
         Section[i].File = fopen (Section[i].Name, "w");
         if (Section[i].File == NULL) {
            fprintf (stderr, "NGCVT: Cannot open file (%s).\n", Section[i].String);
            exit (1);
         }
      }

      /*
      ** Calcualte the hash code for the section selector phrase, and then
      ** increment the count.
      */
      Section[i].HashCode = Hash(Section[i].String);
      i++;
   }

   yyout = Section[0].File;

}     /* InitOutput() */

/* --------------------------------- Hash() -------------------------------- */
/* ------------------------------------------------------------------------- */

unsigned int Hash (char *str)
{
   unsigned int hshval = 0;
   int length = strlen (str);

   while (*str) {
      hshval = ((hshval << 3) ^ tolower(*str)) ^ length;
      str++;
   }

   return (hshval);

}     /* Hash() */

/* --------------------------------- GrowPSP() ----------------------------- */
/* ------------------------------------------------------------------------- */

void GrowPSP (int NewSize)
{
   int   OldSize;
   int  *PSPHdlCnt;
   BYTE *NewHandle;
   BYTE **OldHandle;

   /*
   ** Set the new size of the handle block (in bytes).
   */
   FP_SEG(PSPHdlCnt) = _psp;
   FP_OFF(PSPHdlCnt) = 0x0032;
   OldSize = *PSPHdlCnt;
   *PSPHdlCnt = NewSize;

   /*
   ** Get the pointer to the old handle block.
   */
   FP_SEG(OldHandle) = _psp;
   FP_OFF(OldHandle) = 0x0034;
   NewHandle = calloc (NewSize, 1);

   /*
   ** Allocate the new handles, and copy the first five bytes from the
   ** old handle table to the new handle table.
   */
   memset (NewHandle, 0x00ff, NewSize);
   memcpy (NewHandle, *OldHandle, OldSize);

   /*
   ** Set the PSP to point to the new handle table.
   */
   *OldHandle = NewHandle;

}     /* GrowPSP() */

/* ------------------------------- InitDescList() -------------------------- */
/* ------------------------------------------------------------------------- */
/*
**  Name        : InitDescList
**
**  Description : Initialize the description list.  The list is an allocated
**                array of structures.  Each structre contains a pointer to
**                the item name to recieve the description and the second
**                pointer is to the actual description text.
**
**                The first line in the input file is the number of entries
**                in the description list.  This number is parsed and then
**                used to allocate the array of structures.
**
**  Returns     : None.
**
**  Parameters  : None.
**
*/
void InitDescList(void)
{
   FILE     *iFile;                     /* Input file.                       */
   int      i, j;                       /* Local Indicies.                   */
   char     ItemBuffer[48];             /* Character buffer for item name.   */
   char     DescBuffer[80];             /* Character buffer for description. */
   char     c;                          /* Current input character.          */

   iFile = fopen ("messages.dsc", "rt");
   if (iFile == NULL) {
      fprintf (stderr, "NGCVT: Unable to open input file MESSAGES.DSC.\n");
      return;
   }

   /*
   ** Scan the first line of the file for the number of entries in the
   ** input file.
   */
   fscanf (iFile,"%d\n", &ItemCount);
   if (ItemCount == 0) {
      fprintf (stderr, "NGCVT: Item count not defined or zero in MESSAGES.DSC.\n");
      fclose (iFile);
      return;
   }

   /*
   ** Now allocate the header array.
   */
   HdrList = calloc (ItemCount, sizeof(HEADERS));
   if (HdrList == NULL) {
      fprintf (stderr, "NGCVT: Could not allocate header array for MESSAGES.DSC.\n");
      fclose (iFile);
      return;
   }

   /*
   ** For each item define in the input file, scan the item into the buffers,
   ** and then duplicate the strings into the appropriate fields in the
   ** "HdrList" array.
   */
   for (i=0; i<ItemCount; i++) {
      j = 0;
      fscanf (iFile, "%s ", ItemBuffer);
      HdrList[i].ItemName = strdup(ItemBuffer);
      HdrList[i].HashCode = Hash (ItemBuffer);
      while ((c = (char)getc(iFile)) != '\n') DescBuffer[j++] = c;
      DescBuffer[j] = '\0';
      HdrList[i].Header   = strdup(DescBuffer);
   }

}        /* InitDescList() */

/* ------------------------------- GetDescription() ------------------------ */
/* ------------------------------------------------------------------------- */
/*
**  Name        : GetDescription
**
**  Description : Return a pointer to the short description for the subject
**                identified by the has code passed in as the first parameter.
**
**  Returns     : Pointer to the description string.
**
**  Parameters  : Unsigned integer hash code value for the target description.
**
*/
char * GetDescription (unsigned HashCode, char *Name)
{
   int   i;                             /* Local Index.                      */

   for (i=0; i<ItemCount; i++) {
      if (HdrList[i].HashCode == HashCode) {
         if (strcmp(Name, HdrList[i].ItemName) == 0) {
            return (HdrList[i].Header);
         }
      }
   }

   return (NULL);

}     /* GetDescription() */
